==> klayout-0.15.tar.gz: klayout-0.15/src/strmclip.cc <==

/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/



#include <dbLayout.h>
#include <dbGDS2Writer.h>
#include <dbOASISWriter.h>
#include <dbReader.h>
#include <layParsedLayerSource.h>
#include <tlLog.h>


struct ClipData
{
  ClipData () 
    : file_in (), file_out (), clip_layer (), 
      oasis (false), gzip (false)
  { }

  std::string file_in;
  std::string file_out;
  lay::ParsedLayerSource clip_layer;
  bool oasis;
  bool gzip;
  std::vector <db::DBox> clip_boxes;
  std::string result;
  std::string top;
};


void collect_clip_boxes (const db::Layout &layout, 
                         db::Layout::cell_index_type cell_index, 
                         unsigned int layer, 
                         const db::CplxTrans &trans, 
                         std::vector <db::Box> &boxes)
{
  const db::Cell &cell = layout.cell (cell_index);
  if (! cell.bbox (layer).empty ()) {

    //  any shapes to consider ..
    for (db::ShapeIterator sh = cell.shapes (layer).begin (db::ShapeIterator::All); ! sh.at_end (); ++sh) {
      boxes.push_back (db::Box::from_double (sh->bbox ().transformed (trans)));
    }

    for (db::Cell::const_iterator inst = cell.begin (); ! inst.at_end (); ++inst) {
      for (db::CellInstArray::iterator a = inst->cell_inst ().begin (); ! a.at_end (); ++a) {
        collect_clip_boxes (layout, inst->cell_index (), layer, trans * inst->cell_inst ().complex_trans (*a), boxes);
      }
    }

  }
}

void collect_clip_variants (const db::Layout &layout,
                            db::Layout &target_layout,
                            db::Layout::cell_index_type cell_index,
                            const db::Box &clip_box,
                            std::map <std::pair 
<db::Layout::cell_index_type, db::Box>, db::Layout::cell_index_type> 
&variants)
{
  const db::Cell &cell = layout.cell (cell_index);
  db::box_convert <db::CellInst> bc (layout);

  db::Box cell_box = cell.bbox () & clip_box;
  if (! cell_box.empty ()) {

    std::pair <std::map <std::pair <db::Layout::cell_index_type, db::Box>, db::Layout::cell_index_type>::iterator, bool> vmp = variants.insert (std::make_pair (std::make_pair (cell_index, cell_box), 0));
    if (vmp.second) {
      tl::info << "Create clip variant of cell " << layout.cell_name (cell_index) << ", clip box " << cell_box.to_string ();
      vmp.first->second = target_layout.add_cell (layout.cell_name (cell_index));

      for (db::Cell::touching_iterator inst = cell.begin_touching (cell_box); ! inst.at_end (); ++inst) {
        for (db::CellInstArray::iterator a = inst->cell_inst ().begin_touching (cell_box, bc); ! a.at_end (); ++a) {
          db::Box inst_clip_box = db::Box::from_double (cell_box.transformed (inst->cell_inst ().complex_trans (*a).inverted ()));
          collect_clip_variants (layout, target_layout, inst->cell_index (), inst_clip_box, variants);
}
      }

    }
  }
}

struct EdgeCompareP1Func 
{
  bool operator () (const db::Edge &e1, const db::Edge &e2) const
  {
    return e1.p1 () < e2.p1 ();
  }
};

bool clip_poly (const db::SimplePolygon &poly, const db::Box &box, std::vector <db::SimplePolygon> &clipped_poly)
{
  //  first, extract and sort all edges
  std::vector <db::Edge> edges;
  //  @@@ this is just an approximation; for generic polygon, take holes into account
  edges.reserve (poly.hull ().size ());

  //  create a set of edges to consider
  for (db::SimplePolygon::polygon_edge_iterator e = poly.begin_edge (); ! e.at_end (); ++e) {

    db::Edge edge = *e;

    db::Coord y1 = std::min (edge.p1 ().y (), edge.p2 ().y ());
    db::Coord y2 = std::max (edge.p1 ().y (), edge.p2 ().y ());

    if (y1 < box.p2 ().y () && y2 > box.p1 ().y ()) {

      std::pair <bool, db::Edge> ce = edge.clipped (box);
      if (ce.first) {

        edges.push_back (ce.second);

        db::Edge e1 (db::Point (ce.second.p1 ().x (), edge.p1 ().y ()), ce.second.p1 ());
        db::Edge e2 (ce.second.p2 (), db::Point (ce.second.p2 ().x (), edge.p2 ().y ()));

        if (! e1.is_degenerate () && (ce = e1.clipped (box)).first && ! ce.second.is_degenerate ()) {
          edges.push_back (ce.second);
        }
        if (! e2.is_degenerate () && (ce = e2.clipped (box)).first && ! ce.second.is_degenerate ()) {
          edges.push_back (ce.second);
        }

      } else if (edge.p1 ().x () < box.p1 ().x () || edge.p2 ().x () < box.p1 ().x ()) {

        edge = db::Edge (db::Point (box.p1 ().x (), edge.p1 ().y ()), db::Point (box.p1 ().x (), edge.p2 ().y ()));
        ce = edge.clipped (box);
        if (ce.first) {
          edges.push_back (ce.second);
        }

      } else if (edge.p1 ().x () > box.p2 ().x () || edge.p2 ().x () > box.p2 ().x ()) {

        edge = db::Edge (db::Point (box.p2 ().x (), edge.p1 ().y ()), db::Point (box.p2 ().x (), edge.p2 ().y ()));
        ce = edge.clipped (box);
        if (ce.first) {
          edges.push_back (ce.second);
        }

      }

    }

  }

  //  synthesize horizontal edges at upper and lower boundary of the clip rectangle
  std::vector <std::pair <db::Coord, int> > coord_values;
  std::vector <db::Point> p1_stack;
  std::vector <db::Point> p2_stack;

  for (unsigned int l = 0; l <= 1; ++l) {

    db::Coord y = (l > 0 ? box.p2 ().y () : box.p1 ().y ());

    coord_values.clear ();
    p1_stack.clear ();
    p2_stack.clear ();

    for (std::vector <db::Edge>::const_iterator e = edges.begin (); e != edges.end (); ++e) {
      if (e->p1 ().y () == y) {
        coord_values.push_back (std::make_pair (e->p1 ().x (), -1));
      }
      if (e->p2 ().y () == y) {
        coord_values.push_back (std::make_pair (e->p2 ().x (), 1));
      }
    }

    std::sort (coord_values.begin (), coord_values.end ());

    for (std::vector <std::pair <db::Coord, int> >::const_iterator xv = coord_values.begin (); xv != coord_values.end (); ++xv) {
      db::Point p = db::Point (xv->first, y);
      if (xv->second < 0) {
        if (! p1_stack.empty ()) {
          if (p != p1_stack.back ()) {
            edges.push_back (db::Edge (p1_stack.back (), p));
          }
          p1_stack.pop_back ();
        } else {
          p2_stack.push_back (p);
        }
      } else {
        if (! p2_stack.empty ()) {
          if (p != p2_stack.back ()) {
            edges.push_back (db::Edge (p, p2_stack.back ()));
          }
          p2_stack.pop_back ();
        } else {
          p1_stack.push_back (p);
        }
      }
    }

    tl_assert (p1_stack.empty ());
    tl_assert (p2_stack.empty ());
    coord_values.clear ();

  }

  //  remove all edges being vertical and coincident with the clip box ..
  std::vector <db::Edge>::iterator e_write = edges.begin ();
  for (std::vector <db::Edge>::const_iterator e = edges.begin (); e != edges.end (); ++e) {
    if (e->dx () != 0 || (e->p1 ().x () > box.p1 ().x () && e->p1 ().x () < box.p2 ().x ())) {
      *e_write = *e;
      ++e_write;
    }
  }
  edges.erase (e_write, edges.end ());

  //  and synthesize them again thus removing coincident edges ..
  
  for (unsigned int l = 0; l <= 1; ++l) {

    db::Coord x = (l > 0 ? box.p2 ().x () : box.p1 ().x ());

    coord_values.clear ();
    p1_stack.clear ();
    p2_stack.clear ();

    for (std::vector <db::Edge>::const_iterator e = edges.begin (); e != edges.end (); ++e) {
      if (e->p1 ().x () == x) {
        coord_values.push_back (std::make_pair (e->p1 ().y (), -1));
      }
      if (e->p2 ().x () == x) {
        coord_values.push_back (std::make_pair (e->p2 ().y (), 1));
      }
    }

    std::sort (coord_values.begin (), coord_values.end ());

    for (std::vector <std::pair <db::Coord, int> >::const_iterator xv = coord_values.begin (); xv != coord_values.end (); ++xv) {
      db::Point p = db::Point (x, xv->first);
      if (xv->second < 0) {
        if (! p1_stack.empty ()) {
          if (p != p1_stack.back ()) {
            edges.push_back (db::Edge (p1_stack.back (), p));
          }
          p1_stack.pop_back ();
        } else {
          p2_stack.push_back (p);
        }
      } else {
        if (! p2_stack.empty ()) {
          if (p != p2_stack.back ()) {
            edges.push_back (db::Edge (p, p2_stack.back ()));
          }
          p2_stack.pop_back ();
        } else {
          p1_stack.push_back (p);
        }
      }
    }

    tl_assert (p1_stack.empty ());
    tl_assert (p2_stack.empty ());
    coord_values.clear ();

  }

  //  create polygons from the edge sets
  std::sort (edges.begin (), edges.end (), EdgeCompareP1Func ());

  std::vector <db::Point> ctr;

  bool any = true;
  while (any) {

    any = false;
    std::vector <db::Edge>::iterator oe;
    for (oe = edges.begin (); oe != edges.end (); ++oe) {
      if (! oe->is_degenerate ()) {
        any = true;
        break;
      }
    }

    if (any) {

      ctr.clear ();

      bool found = true;
      while (found) {

        db::Edge curr_e (*oe);

        ctr.push_back (curr_e.p1 ());

        *oe = db::Edge (curr_e.p1 (), curr_e.p1 ()); //  marks "used already"

        double smin = 0.0;
        found = false;

        for (std::vector <db::Edge>::iterator oe_next = std::lower_bound (edges.begin (), edges.end (), db::Edge (curr_e.p2 (), curr_e.p2 ()), EdgeCompareP1Func ());
             oe_next != edges.end () && oe_next->p1 () == curr_e.p2 (); ++oe_next) {

          if (! oe_next->is_degenerate ()) {
            double s = double (db::sprod (curr_e, *oe_next)) / (curr_e.double_length () * oe_next->double_length ()) - 1;
            if (db::vprod_sign (curr_e, *oe_next) >= 0) {
              s = -s;
            }
            if (s < smin || !found) {
              smin = s;
              oe = oe_next;
            }
            found = true;
          } 

        }

      }

      clipped_poly.push_back (db::SimplePolygon ());
      clipped_poly.back ().assign_hull (ctr.begin (), ctr.end ());

    }

  }

  return true;

}

void clip_cell (const db::Layout &layout, 
                db::Layout::cell_index_type cell_index, 
                db::Layout &target_layout, 
                db::Layout::cell_index_type target_cell_index, 
                const db::Box &clip_box,
                std::map <std::pair <db::Layout::cell_index_type, db::Box>, db::Layout::cell_index_type> &variants)
{
  const db::Cell &cell = layout.cell (cell_index);
  db::Cell &target_cell = target_layout.cell (target_cell_index);

  if (cell.bbox ().inside (clip_box)) {

    //  simplification of shape copy op in case of no clipping ..
    for (unsigned int l = 0; l < layout.layers (); ++l) {
      if (layout.is_valid_layer (l)) {
        target_cell.shapes (l).translate (cell.shapes (l), target_layout.shape_repository ());
      }
    }

  } else {

    for (unsigned int l = 0; l < layout.layers (); ++l) {
      
      if (layout.is_valid_layer (l)) {

        for (db::ShapeIterator sh = cell.shapes (l).begin_touching (clip_box, db::ShapeIterator::All); ! sh.at_end (); ++sh) {

          if (sh->is_box ()) {

            db::Box new_box = sh->box () & clip_box;
            if (! new_box.empty ()) {
              if (sh->has_prop_id ()) {
                target_cell.shapes (l).basic_insert (db::BoxWithProperties (new_box, sh->prop_id ()));
              } else {
                target_cell.shapes (l).insert (new_box);
              }
            }

          } else if (sh->is_path () && sh->bbox ().inside (clip_box)) {

            db::Path path;
            sh->path (path);

            if (sh->has_prop_id ()) {
              target_cell.shapes (l).insert (db::PathRefWithProperties (db::PathRef (path, target_layout.shape_repository ()), sh->prop_id ()));
            } else {
              target_cell.shapes (l).insert (db::PathRef (path, target_layout.shape_repository ()));
            }

          } else if (sh->is_simple_polygon () || sh->is_path ()) {

            db::SimplePolygon poly;

            if (sh->is_path ()) {
              db::Path path;
              sh->path (path);
              poly = path.simple_polygon ();
            } else {
              sh->simple_polygon (poly);
            }

            std::vector <db::SimplePolygon> clipped_polygons;

            if (! poly.box ().inside (clip_box) && clip_poly (poly, clip_box, clipped_polygons)) {
              for (std::vector <db::SimplePolygon>::const_iterator cp = clipped_polygons.begin (); cp != clipped_polygons.end (); ++cp) {
                if (sh->has_prop_id ()) {
                  target_cell.shapes (l).insert (db::SimplePolygonRefWithProperties (db::SimplePolygonRef (*cp, target_layout.shape_repository ()), sh->prop_id ()));
                } else {
                  target_cell.shapes (l).insert (db::SimplePolygonRef (*cp, target_layout.shape_repository ()));
                }
              }
            } else {
              if (sh->has_prop_id ()) {
                target_cell.shapes (l).insert (db::SimplePolygonRefWithProperties (db::SimplePolygonRef (poly, target_layout.shape_repository ()), sh->prop_id ()));
              } else {
                target_cell.shapes (l).insert (db::SimplePolygonRef (poly, target_layout.shape_repository ()));
              }
            }

          } else if (sh->is_text ()) {

            if (sh->bbox ().inside (clip_box)) {
              db::Text text;
              sh->text (text);
              if (sh->has_prop_id ()) {
                target_cell.shapes (l).insert (db::TextRefWithProperties (db::TextRef (text, target_layout.shape_repository ()), sh->prop_id ()));
              } else {
                target_cell.shapes (l).insert (db::TextRef (text, target_layout.shape_repository ()));
              }
            }

          } else {
            tl_assert (false); // invalid shape type encountered
          }

        }

      }

    }

  }

  db::box_convert <db::CellInst> bc (layout);

  for (db::Cell::touching_iterator inst = cell.begin_touching (clip_box); ! inst.at_end (); ++inst) {

    if (inst->cell_inst ().bbox (bc).inside (clip_box)) {

      //  instance is completely inside
      db::CellInstArray new_inst = inst->cell_inst ();

      const db::Cell &inst_cell = layout.cell (inst->cell_index ());

      std::map <std::pair <db::Layout::cell_index_type, db::Box>, db::Layout::cell_index_type>::const_iterator vmp = variants.find (std::make_pair (inst->cell_index (), inst_cell.bbox ()));
      tl_assert (vmp != variants.end ());

      new_inst.object () = db::CellInst (vmp->second);

      //  TODO: keep properties 
      target_cell.insert (new_inst);

    } else {

      for (db::CellInstArray::iterator a = inst->cell_inst ().begin_touching (clip_box, bc); ! a.at_end (); ++a) {

        db::Box inst_clip_box = db::Box::from_double (clip_box.transformed (inst->cell_inst ().complex_trans (*a).inverted ()));
        const db::Cell &inst_cell = layout.cell (inst->cell_index ());

        inst_clip_box &= inst_cell.bbox ();

        if (! inst_clip_box.empty ()) {

          std::map <std::pair <db::Layout::cell_index_type, db::Box>, db::Layout::cell_index_type>::const_iterator vmp = variants.find (std::make_pair (inst->cell_index (), inst_clip_box));
          tl_assert (vmp != variants.end ());

          db::CellInstArray new_inst;

          if (inst->is_complex ()) {
            new_inst = db::CellInstArray (db::CellInst (vmp->second), inst->cell_inst ().complex_trans (*a));
          } else {
            new_inst = db::CellInstArray (db::CellInst (vmp->second), *a);
          }

          target_cell.insert (new_inst);

        }

      }

    }

  }

}

void clip (const ClipData &data)
{
  db::Manager m;
  db::Layout layout (&m);
  db::Layout target_layout (&m);
  const db::CplxTrans trans;

  {
    tl::InputZLibFile stream (data.file_in);
    db::Reader reader (stream);
    reader.read (layout);
  }

  //  create the layers in the target layout as well
  for (unsigned int i = 0; i < layout.layers (); ++i) {
    if (layout.is_valid_layer (i)) {
      target_layout.insert_layer (i, layout.get_properties (i));
    }
  }

  //  copy the properties repository in order to have the same ID mapping
  target_layout.properties_repository () = layout.properties_repository ();
  target_layout.dbu (layout.dbu ());

  //  look for the clip layer
  int clip_layer_index = -1;
  for (unsigned int i = 0; i < layout.layers (); ++i) {
    if (layout.is_valid_layer (i) && data.clip_layer.match (layout.get_properties (i))) {
      clip_layer_index = int (i);
      break;
    }
  }

  tl::info << "Clip layer index is " << clip_layer_index;

  //  get top cells
  std::vector <db::CellGraph::cell_index_type> top_cells;
  if (data.top.empty ()) {
    top_cells.assign (layout.begin_top_down (), layout.end_top_cells ());
  } else {
    std::pair<bool, db::CellGraph::cell_index_type> tc = layout.cell_by_name (data.top.c_str ());
    if (! tc.first) {
      throw tl::Exception ("Cell %s is not a valid cell in the input layout", data.top.c_str ());
    }
    top_cells.push_back (tc.second);
  }

  //  go through the top cells
  for (std::vector <db::CellGraph::cell_index_type>::const_iterator tc = top_cells.begin (); tc != top_cells.end (); ++tc) {

    std::vector <db::Box> clip_boxes;

    //  add the explicit boxes first
    for (std::vector <db::DBox>::const_iterator b = data.clip_boxes.begin (); b != data.clip_boxes.end (); ++b) {
      clip_boxes.push_back (db::Box::from_double (*b * (1.0 / layout.dbu ())));
    }

    //  fetch the boxes of the clip shapes
    if (clip_layer_index >= 0) {
      collect_clip_boxes (layout, *tc, clip_layer_index, trans, clip_boxes);
    }

    //  sort our duplicate boxes
    std::sort (clip_boxes.begin (), clip_boxes.end ());
    clip_boxes.erase (std::unique (clip_boxes.begin (), clip_boxes.end ()), clip_boxes.end ());

    tl::info << "Clip boxes are:";
    for (std::vector <db::Box>::const_iterator cbx = clip_boxes.begin (); cbx != clip_boxes.end (); ++cbx) {
      tl::info << "  " << cbx->to_string ();
    }

    //  create clip variants of cells
    std::map <std::pair <db::Layout::cell_index_type, db::Box>, db::Layout::cell_index_type> variants;
    for (std::vector <db::Box>::const_iterator cbx = clip_boxes.begin (); cbx != clip_boxes.end (); ++cbx) {
      collect_clip_variants (layout, target_layout, *tc, *cbx, variants);
    }

    //  actually do the clipping by filling the variants
    for (std::map <std::pair <db::Layout::cell_index_type, db::Box>, db::Layout::cell_index_type>::const_iterator var = variants.begin (); var != variants.end (); ++var) {
      clip_cell (layout, var->first.first, target_layout, var->second, var->first.second, variants);
    }

    //  create "very top" cells to put the result cells into
    std::string result_top;
    if (! data.result.empty ()) {
      result_top = data.result;
    } else {
      result_top = std::string ("CLIPPED_") + layout.cell_name (*tc);
    }
    db::Layout::cell_index_type clip_top = target_layout.add_cell (result_top.c_str ());
    db::Cell &clip_top_cell = target_layout.cell (clip_top);

    for (std::map <std::pair <db::Layout::cell_index_type, db::Box>, db::Layout::cell_index_type>::const_iterator var = variants.begin (); var != variants.end (); ++var) {
      if (var->first.first == *tc) {
        clip_top_cell.insert (db::CellInstArray (db::CellInst (var->second), db::Trans ()));
      }
    }

  }

  //  write the layout
  tl::OutputStreamBase *out_file = 0;
  try {

    if (data.gzip) {
      out_file = new tl::OutputZLibFile (data.file_out);
    } else {
      out_file = new tl::OutputFile (data.file_out);
    }

    tl::OutputStream stream (*out_file);

    if (data.oasis) {
      db::OASISWriter writer;
      writer.write (target_layout, stream, db::SaveLayoutOptions ());
    } else {
      db::GDS2Writer writer;
      writer.write (target_layout, stream, db::SaveLayoutOptions ());
    }

    delete out_file;

  } catch (...) {
    if (out_file) {
      delete out_file;
    }
    throw;
  }
}

void print_syntax ()
{
  printf ("Syntax: strmclip [<options>] <infile> <outfile>\n");
  printf ("\n");
  printf ("Options are:\n");
  printf ("  -l 'l/d'      take clip regions from layer l, datatype d\n");
  printf ("  -o            produce oasis output\n");
  printf ("  -g            produce gds output\n");
  printf ("  -z            gzip output\n");
  printf ("  -t 'cell'     use this cell from input (default: determine top cell automatically)\n");
  printf ("  -x 'name'     use this cell as top cell in output\n");
  printf ("  -r 'l,b,r,t'  explicitly specify a clip retangle (can be present multiple times)\n");
}

int 
main (int argc, char *argv [])
{
  try {

    ClipData data;

    for (int n = 1; n < argc; ++n) {

      if (std::string (argv [n]) == "-h") {
        print_syntax ();
        return 0;
      } else if (std::string (argv [n]) == "-o") {
        data.oasis = true;
      } else if (std::string (argv [n]) == "-g") {
        data.oasis = false;
      } else if (std::string (argv [n]) == "-z") {
        data.gzip = true;
      } else if (std::string (argv [n]) == "-x") {
        if (n < argc + 1) {
          ++n;
          data.result = argv [n];
        } 
      } else if (std::string (argv [n]) == "-t") {
        if (n < argc + 1) {
          ++n;
          data.top = argv [n];
        } 
      } else if (std::string (argv [n]) == "-r") {
        if (n < argc + 1) {
          ++n;
          tl::Extractor ex (argv [n]);
          double l = 0.0, b = 0.0, r = 0.0, t = 0.0;
          ex.read (l); ex.expect (",");
          ex.read (b); ex.expect (",");
          ex.read (r); ex.expect (",");
          ex.read (t); ex.expect_end ();
          data.clip_boxes.push_back (db::DBox (l, b, r, t));
        } 
      } else if (std::string (argv [n]) == "-l") {
        if (n < argc + 1) {
          ++n;
          data.clip_layer = lay::ParsedLayerSource (argv [n]);
        } 
      } else if (argv [n][0] == '-') {
        print_syntax ();
        throw tl::Exception ("Unknown option: " + std::string (argv [n]));
      } else if (data.file_in.empty ()) {
        data.file_in = argv [n];
      } else if (data.file_out.empty ()) {
        data.file_out = argv [n];
      } else {
        print_syntax ();
        throw tl::Exception ("Superfluous command element: " + std::string (argv [n]));
      }

    }

    if (data.file_in.empty () || data.file_out.empty ()) {
      print_syntax ();
      throw tl::Exception ("Input or output file name missing");
    }

    clip (data);

  } catch (std::exception &ex) {
    tl::error << ex.what ();
    return 1;
  } catch (tl::Exception &ex) {
    tl::error << ex.msg ();
    return 1;
  } catch (...) {
    tl::error << "unspecific error";
    return 1;
  }

  return 0;
}

==> klayout-0.19.3.tar.gz: klayout-0.19.3/src/strmclip.cc <==

/*

  KLayout Layout Viewer
  Copyright (C) 2006-2007 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/



#include "dbClip.h"
#include "dbLayout.h"
#include "dbGDS2Writer.h"
#include "dbOASISWriter.h"
#include "dbReader.h"
#include "layParsedLayerSource.h"
#include "tlLog.h"


struct ClipData
{
  ClipData () 
    : file_in (), file_out (), clip_layer (), 
      oasis (false), gzip (false)
  { }

  std::string file_in;
  std::string file_out;
  lay::ParsedLayerSource clip_layer;
  bool oasis;
  bool gzip;
  std::vector <db::DBox> clip_boxes;
  std::string result;
  std::string top;
};


void clip (const ClipData &data)
{
  db::Manager m;
  db::Layout layout (&m);
  db::Layout target_layout (&m);

  {
    tl::InputZLibFile stream (data.file_in);
    db::Reader reader (stream);
    reader.read (layout);
  }

  //  create the layers in the target layout as well
  for (unsigned int i = 0; i < layout.layers (); ++i) {
    if (layout.is_valid_layer (i)) {
      target_layout.insert_layer (i, layout.get_properties (i));
    }
  }

  //  copy the properties repository in order to have the same ID mapping
  target_layout.properties_repository () = layout.properties_repository ();
  target_layout.dbu (layout.dbu ());

  //  look for the clip layer
  int clip_layer_index = -1;
  for (unsigned int i = 0; i < layout.layers (); ++i) {
    if (layout.is_valid_layer (i) && data.clip_layer.match (layout.get_properties (i))) {
      clip_layer_index = int (i);
      break;
    }
  }

  tl::info << "Clip layer index is " << clip_layer_index;

  //  get top cells
  std::vector <db::cell_index_type> top_cells;
  if (data.top.empty ()) {
    top_cells.assign (layout.begin_top_down (), layout.end_top_cells ());
  } else {
    std::pair<bool, db::cell_index_type> tc = layout.cell_by_name (data.top.c_str ());
    if (! tc.first) {
      throw tl::Exception ("Cell %s is not a valid cell in the input layout", data.top.c_str ());
    }
    top_cells.push_back (tc.second);
  }

  //  go through the top cells
  for (std::vector <db::cell_index_type>::const_iterator tc = top_cells.begin (); tc != top_cells.end (); ++tc) {

    std::vector <db::Box> clip_boxes;

    //  add the explicit boxes first
    for (std::vector <db::DBox>::const_iterator b = data.clip_boxes.begin (); b != data.clip_boxes.end (); ++b) {
      clip_boxes.push_back (db::Box::from_double (*b * (1.0 / layout.dbu ())));
    }

    //  fetch the boxes of the clip shapes
    if (clip_layer_index >= 0) {
      collect_clip_boxes (layout, *tc, clip_layer_index, clip_boxes);
    }

    //  sort our duplicate boxes
    std::sort (clip_boxes.begin (), clip_boxes.end ());
    clip_boxes.erase (std::unique (clip_boxes.begin (), clip_boxes.end ()), clip_boxes.end ());

    tl::info << "Clip boxes are:";
    for (std::vector <db::Box>::const_iterator cbx = clip_boxes.begin (); cbx != clip_boxes.end (); ++cbx) {
      tl::info << "  " << cbx->to_string ();
    }

    std::vector<db::cell_index_type> new_cells = db::clip_layout (layout, target_layout, *tc, clip_boxes);

    //  create "very top" cells to put the result cells into
    std::string result_top;
    if (! data.result.empty ()) {
      result_top = data.result;
    } else {
      result_top = std::string ("CLIPPED_") + layout.cell_name (*tc);
    }
    db::cell_index_type clip_top = target_layout.add_cell (result_top.c_str ());
    db::Cell &clip_top_cell = target_layout.cell (clip_top);

    for (std::vector <db::cell_index_type>::const_iterator cc = new_cells.begin (); cc != new_cells.end (); ++cc) {
      clip_top_cell.insert (db::CellInstArray (db::CellInst (*cc), db::Trans ()));
    }

  }

  //  write the layout
  tl::OutputStreamBase *out_file = 0;
  try {

    if (data.gzip) {
      out_file = new tl::OutputZLibFile (data.file_out);
    } else {
      out_file = new tl::OutputFile (data.file_out);
    }

    tl::OutputStream stream (*out_file);

    if (data.oasis) {
      db::OASISWriter writer;
      writer.write (target_layout, stream, db::SaveLayoutOptions ());
    } else {
      db::GDS2Writer writer;
      writer.write (target_layout, stream, db::SaveLayoutOptions ());
    }

    delete out_file;

  } catch (...) {
    if (out_file) {
      delete out_file;
    }
    throw;
  }
}

void print_syntax ()
{
  printf ("Syntax: strmclip [<options>] <infile> <outfile>\n");
  printf ("\n");
  printf ("Options are:\n");
  printf ("  -l 'l/d'      take clip regions from layer l, datatype d\n");
  printf ("  -o            produce oasis output\n");
  printf ("  -g            produce gds output\n");
  printf ("  -z            gzip output\n");
  printf ("  -t 'cell'     use this cell from input (default: determine top cell automatically)\n");
  printf ("  -x 'name'     use this cell as top cell in output\n");
  printf ("  -r 'l,b,r,t'  explicitly specify a clip retangle (can be present multiple times)\n");
}

int 
main (int argc, char *argv [])
{
  try {

    ClipData data;

    for (int n = 1; n < argc; ++n) {

      if (std::string (argv [n]) == "-h") {
        print_syntax ();
        return 0;
      } else if (std::string (argv [n]) == "-o") {
        data.oasis = true;
      } else if (std::string (argv [n]) == "-g") {
        data.oasis = false;
      } else if (std::string (argv [n]) == "-z") {
        data.gzip = true;
      } else if (std::string (argv [n]) == "-x") {
        if (n < argc + 1) {
          ++n;
          data.result = argv [n];
        } 
      } else if (std::string (argv [n]) == "-t") {
        if (n < argc + 1) {
          ++n;
          data.top = argv [n];
        } 
      } else if (std::string (argv [n]) == "-r") {
        if (n < argc + 1) {
          ++n;
          tl::Extractor ex (argv [n]);
          double l = 0.0, b = 0.0, r = 0.0, t = 0.0;
          ex.read (l); ex.expect (",");
          ex.read (b); ex.expect (",");
          ex.read (r); ex.expect (",");
          ex.read (t); ex.expect_end ();
          data.clip_boxes.push_back (db::DBox (l, b, r, t));
        } 
      } else if (std::string (argv [n]) == "-l") {
        if (n < argc + 1) {
          ++n;
          data.clip_layer = lay::ParsedLayerSource (argv [n]);
        } 
      } else if (argv [n][0] == '-') {
        print_syntax ();
        throw tl::Exception ("Unknown option: " + std::string (argv [n]));
      } else if (data.file_in.empty ()) {
        data.file_in = argv [n];
      } else if (data.file_out.empty ()) {
        data.file_out = argv [n];
      } else {
        print_syntax ();
        throw tl::Exception ("Superfluous command element: " + std::string (argv [n]));
      }

    }

    if (data.file_in.empty () || data.file_out.empty ()) {
      print_syntax ();
      throw tl::Exception ("Input or output file name missing");
    }

    clip (data);

  } catch (std::exception &ex) {
    tl::error << ex.what ();
    return 1;
  } catch (tl::Exception &ex) {
    tl::error << ex.msg ();
    return 1;
  } catch (...) {
    tl::error << "unspecific error";
    return 1;
  }

  return 0;
}
